/*---------------------------------------------------------------------------------
 bbEdgeFlip ( 2004 Slade Taylor [bladestaylor@yahoo.com])
 ----------------------------------------------------------------------------------
 bbEdgeFlip is a plugin for Blackbox for Windows.  For more information,
 please visit [http://bb4win.org] or [http://sourceforge.net/projects/bb4win].
 ----------------------------------------------------------------------------------
 bbEdgeFlip is free software, released under the GNU General Public License,
 version 2, as published by the Free Software Foundation.  It is distributed
 WITHOUT ANY WARRANTY--without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  Please see the GNU General Public
 License for more details:  [http://www.fsf.org/licenses/gpl.html].
 --------------------------------------------------------------------------------*/

#include "bbEdgeFlip.h"

//---------------------------------------------------------------------------------

int beginPlugin(HINSTANCE hPluginInstance)
{
    WNDCLASS wc;
    ZeroMemory(&wc, sizeof(wc));
    wc.lpfnWndProc = PluginProc;
    wc.hInstance = hPluginInstance;
    wc.lpszClassName = pluginInfo(APPNAME);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    RegisterClass(&wc);

    CreateFlipWindow(hPluginInstance, &hLeftWnd);
    CreateFlipWindow(hPluginInstance, &hRightWnd);
    SendMessage(hCoreWnd = GetBBWnd(), BB_REGISTERMESSAGE, (WPARAM)hRightWnd, (LPARAM)bb_messages);

    *(FARPROC*)&pSetLayeredWindowAttributes =
        GetProcAddress(GetModuleHandle("user32.dll"), "SetLayeredWindowAttributes");

    *(FARPROC*)&pGetDesktopInfo =
        GetProcAddress(GetModuleHandle("blackbox.exe"), "GetDesktopInfo");

    int i = GetModuleFileName(hPluginInstance, RCpath, MAX_LINE_LENGTH);

    char hook_path[MAX_LINE_LENGTH];
    strcpy(hook_path, RCpath);

    if (i > 3) { RCpath[i - 3] = 'r'; RCpath[i - 2] = 'c'; RCpath[i - 1] = 0; }

    sprintf(hook_path + i - 14, "%s", "WMHook.dll");
    if (pGetDesktopInfo && (hWinMoveHook = LoadLibrary(hook_path)))
    {
        *(FARPROC*)&pSetHook = GetProcAddress(hWinMoveHook, "SetHook");
        *(FARPROC*)&pUnsetHook = GetProcAddress(hWinMoveHook, "UnsetHook");
        if (pSetHook && pUnsetHook)
            pSetHook();
    }

    Reconfig();
    return 0;
}

//---------------------------------------------------------------------------------

void endPlugin(HINSTANCE hPluginInstance)
{
    DestroyFlipWindow(hLeftWnd);
    DestroyFlipWindow(hRightWnd);
    SendMessage(hCoreWnd, BB_UNREGISTERMESSAGE, (WPARAM)hRightWnd, (LPARAM)bb_messages);
    UnregisterClass(pluginInfo(APPNAME), hPluginInstance);

    if (pUnsetHook)
        pUnsetHook();

    if (hWinMoveHook)
        FreeLibrary(hWinMoveHook);
}

//---------------------------------------------------------------------------------

LRESULT CALLBACK PluginProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        default:
            return DefWindowProc(hwnd, message, wParam, lParam);

        case WM_PAINT:
            {
                RECT r;
                PAINTSTRUCT ps;
                HBRUSH h_brush = CreateSolidBrush(BrushColor);
                HDC hdcMain = BeginPaint(hwnd, &ps);
                GetClientRect(hwnd, &r);
                FillRect(hdcMain, &r, h_brush);
                DeleteObject(h_brush);
                EndPaint(hwnd, &ps);
            }
            break;

        case WM_MOUSEMOVE:
            if (bControlFlip)
            {
                UINT interval = 50;
                bControlFlip = false;
                if (bTriggerType == TRIGGER_SLIDE)
                    GetCursorPos(&InitialMousePoint);
                else
                    interval = ((GetAsyncKeyState(VK_CONTROL) & 0x8000) != 0) ? 0: InitialDelay;
                SetTimer(hwnd, FLIP_TIMER, interval, NULL);
                SetTimer(hwnd, DCHECK_TIMER, 50, NULL);
            }
            break;

        case WM_TIMER:
            GetCursorPos(&MouseEventPoint);
            switch (wParam)
            {
                case DCHECK_TIMER:
                    if ((MouseEventPoint.x > RightEdgeThres) ||
                        (MouseEventPoint.x < LeftEdgeThres)) break;
                    goto kill_tm;

                case FLIP_TIMER:
                    if (ProcessFlip(
                        hwnd,
                        hRightWnd,
                        DESK_RIGHT,
                        MouseEventPoint.x > RightEdgeThres,
                        MouseEventPoint.x < LeftEdgeThres)) break;
                    if (ProcessFlip(
                        hwnd,
                        hLeftWnd,
                        DESK_LEFT,
                        MouseEventPoint.x < LeftEdgeThres,
                        MouseEventPoint.x > RightEdgeThres)) break;
kill_tm:            KillTimer(hwnd, DCHECK_TIMER);
                    KillTimer(hwnd, FLIP_TIMER);
                    bControlFlip = true;
            }
            break;

        case FLIP_WM_MOVING:
            GetCursorPos(&MouseEventPoint);
            if (MouseEventPoint.x < (VScreenX + 3))
                ProcessWindowMove((HWND)wParam, 6, VScreenX + VScreenWidth - 1);
            else if (MouseEventPoint.x > (VScreenX + VScreenWidth - 3))
                ProcessWindowMove((HWND)wParam, 7, VScreenX);
            else
                bControlMove = false;
            break;

        case BB_RECONFIGURE:
            Reconfig();
    }
    return 0;
}

//---------------------------------------------------------------------------------

bool ProcessFlip(HWND hwnd, HWND hwnd_test, WPARAM wParam, bool test_before, bool test_after)
{
    int new_x;
    if (hwnd == hwnd_test)
    {
        if (test_before)
        {
            if (bTriggerType == TRIGGER_SLIDE)
            {
                long test = (MouseEventPoint.y - InitialMousePoint.y);
                if (((test < 0) ? -test: test) > SlideThres)
                {
                    InitialMousePoint.y = MouseEventPoint.y;
                    if (bActionType == ACTION_WINDOWS)
                    {
                        InitialMousePoint.x = MouseEventPoint.x;
                        goto set_2;
                    }
                    else //...if (bActionType == ACTION_UNIX)
                    {
                        InitialMousePoint.x =
                            new_x = VScreenX + ((hwnd == hRightWnd) ? 100: (VScreenWidth - 100));
                        goto set_1;
                    }
                }
            }
            else if (bActionType == ACTION_WINDOWS)
            {
                SetTimer(hwnd, FLIP_TIMER, RepeatDelay, NULL);
                goto set_2;
            }
            else //...if (bActionType == ACTION_UNIX)
            {
                new_x = VScreenX + ((hwnd == hRightWnd) ? 0: VScreenWidth);
                SetTimer(hwnd, FLIP_TIMER, 50, NULL);
set_1:          SetCursorPos(new_x, MouseEventPoint.y);
set_2:          PostMessage(hCoreWnd, BB_WORKSPACE, wParam, NULL);
            }
            return true;
        }
        return test_after;
    }
    return false;
}

//---------------------------------------------------------------------------------

void ProcessWindowMove(HWND h, WPARAM wp, int x)
{
    if (!bControlMove)
    {
        bool bSticky = CheckSticky(h);
        if (!bSticky)
            MakeSticky(h);
        SendMessage(hCoreWnd, BB_WORKSPACE, wp, (LPARAM)h);
        pGetDesktopInfo(&CurrentDesktopInfo);
        SetTaskWorkspace(h, CurrentDesktopInfo.number);
        SetCursorPos(x, MouseEventPoint.y);
        if (!bSticky)
            RemoveSticky(h);
        bControlMove = true;
    }
}

//---------------------------------------------------------------------------------

void Reconfig()
{
    RECT r;
    RCSettings();

    int screen_width = GetSystemMetrics(SM_CXSCREEN);
    int screen_height = GetSystemMetrics(SM_CYSCREEN);
    int edge_height =
        VScreenHeight = GetSystemMetrics(SM_CYVIRTUALSCREEN);
    VScreenWidth = GetSystemMetrics(SM_CXVIRTUALSCREEN);

    if ((VScreenWidth > screen_width) || (VScreenHeight > screen_height))
    {
        VScreenX = GetSystemMetrics(SM_XVIRTUALSCREEN);
        VScreenY = GetSystemMetrics(SM_YVIRTUALSCREEN);
    }
    else
    {
        VScreenX = 0;
        VScreenWidth  = screen_width;
        VScreenHeight = screen_height;
        SystemParametersInfo(SPI_GETWORKAREA, NULL, &r, NULL);
        edge_height = r.bottom - r.top;
        VScreenY = r.top;
    }

    LeftEdgeThres = VScreenX + EdgeWidth;
    RightEdgeThres = VScreenX + VScreenWidth - EdgeWidth;
    SlideThres = SlideThresPercent * VScreenHeight / 100;

    ReconfigFlipWindow(hLeftWnd, VScreenX, LeftEdgeAlignment, LeftEdgeHeightPercent, edge_height);
    ReconfigFlipWindow(hRightWnd, RightEdgeThres, RightEdgeAlignment, RightEdgeHeightPercent, edge_height);
}

//---------------------------------------------------------------------------------

void ReconfigFlipWindow(HWND hwnd, int new_x, char *alignment, int height_percent, int edge_height)
{
    int area_height = (height_percent < 0) ? edge_height: (height_percent * VScreenHeight / 100);

    SetWindowPos(
        hwnd,
        HWND_TOPMOST,
        new_x,
        GetEdgeTop(alignment, height_percent, area_height),
        EdgeWidth,
        area_height,
        SWP_NOACTIVATE);

    if (bTrans && pSetLayeredWindowAttributes)
    {
        SetWindowLongPtr(hwnd, GWL_EXSTYLE, WS_EX_TOOLWINDOW|WS_EX_LAYERED);
        pSetLayeredWindowAttributes(hwnd, NULL, 1, LWA_ALPHA);
    }
    else
        SetWindowLongPtr(hwnd, GWL_EXSTYLE, WS_EX_TOOLWINDOW);

    InvalidateRect(hwnd, NULL, false);
}

//---------------------------------------------------------------------------------

int GetEdgeTop(char *alignment, int height_percent, int height_pix)
{
    return  (height_percent < 0) ? VScreenY:
            !strcmp(alignment, "Center") ? (VScreenHeight - height_pix) / 2:
            !strcmp(alignment, "Bottom") ? (VScreenHeight - height_pix):
            VScreenY;
}

//---------------------------------------------------------------------------------

void CreateFlipWindow(HINSTANCE hinstance, HWND *hwnd)
{
    *hwnd = CreateWindowEx(
        WS_EX_TOOLWINDOW,
        pluginInfo(APPNAME),
        NULL,
        WS_POPUP|WS_VISIBLE,
        0, 0, 0, 0,
        NULL,
        NULL,
        hinstance,
        NULL);
    MakeSticky(*hwnd);
}

//---------------------------------------------------------------------------------

void DestroyFlipWindow(HWND hwnd)
{
    RemoveSticky(hwnd);
    DestroyWindow(hwnd);
}

//---------------------------------------------------------------------------------

void RCSettings()
{
    char ActionText[8], TriggerText[6];

    RC_String("Action:",                ActionText,                 "Unix",     7);
    RC_String("Trigger:",               TriggerText,                "Delay",    5);
    RC_Int("Slide.Distance:",           &SlideThresPercent,         30);
    RC_Int("Delay.Initial:",            &InitialDelay,              250);
    RC_Int("Delay.Repeat:",             &RepeatDelay,               500);
    RC_Int("Width:",                    &EdgeWidth,                 2);
    RC_String("Left.Align:",            LeftEdgeAlignment,          "Top",      6);
    RC_Int("Left.Height:",              &LeftEdgeHeightPercent,     -1);
    RC_String("Right.Align:",           RightEdgeAlignment,         "Top",      6);
    RC_Int("Right.Height:",             &RightEdgeHeightPercent,    -1);
    RC_Bool("Transparent:",             &bTrans,                    true);
    RC_Color("Color:",                  &BrushColor,                "#333333");

    bActionType = (0 != _stricmp(ActionText, "Windows"));
    bTriggerType = (0 != _stricmp(TriggerText, "Slide"));
}

//---------------------------------------------------------------------------------

void RC_Int(char *label, int *target, int default_val)
{
    *target = ReadInt(RCpath, label, default_val);
    WriteInt(RCpath, label, *target);
}

//---------------------------------------------------------------------------------

void RC_Bool(char *label, bool *target, bool default_val)
{
    *target = ReadBool(RCpath, label, default_val);
    WriteBool(RCpath, label, *target);
}

//---------------------------------------------------------------------------------

void RC_Color(char *label, COLORREF *target, char *default_val)
{
    *target = ReadColor(RCpath, label, default_val);
    WriteColor(RCpath, label, *target);
}

//---------------------------------------------------------------------------------

void RC_String(char *label, char *target, char *default_val, int max_len)
{
    strncpy(target, ReadString(RCpath, label, default_val), max_len);
    target[max_len] = 0;
    WriteString(RCpath, label, target);
}

//---------------------------------------------------------------------------------
